proc Help { tpl msg target } {
    global HELP
    set w $tpl.help
    if [winfo exists $w] { destroy $w }
    Unset HELP
    toplevel $w
    wm transient $w $tpl
    wm withdraw $w
    wm title $w ""
    
    # MainFrame ####################################################################
    set mf [MainFrame $w.mf -textvariable HELP(status) -progressvar HELP(progress) -progresstype infinite -progressmax 100]
    packef $mf
    $mf showstatusbar progression
    set tb [$mf addtoolbar]
    set mff [$mf getframe]
    
    # ToolBar ######################################################################
    pack [ButtonBox $tb.bb1 -spacing 0] -side left
    $tb.bb1 add -image printer -helptext "  <F4> " -relief link -takefocus 0 \
        -highlightthickness 0 -command ""
    pack [Separator $tb.sep1 -orient vertical -relief groove] -side left -fill y -padx 2
    pack [ButtonBox $tb.bb2 -spacing 0 ] -side right
    $tb.bb2 add -image quit -helptext "  <Esc> " -relief link -takefocus 0 \
        -highlightthickness 0 -command "Escape $w HELP"

    # text #########################################################################
    pack [label $mff.l -text " $msg " -bg grey60 -fg yellow] -expand yes -fill x
    pack [frame $mff.f] -expand yes -fill both
    pack [scrollbar $mff.f.sbx -orient horizontal -command "$mff.f.txt xview"] \
        -side bottom -fill x
    pack [scrollbar $mff.f.sby -command "$mff.f.txt yview"] -side right -fill y
    pack [text $mff.f.txt -wrap none -yscrollcommand "$mff.f.sby set" \
            -xscrollcommand "$mff.f.sbx set"\
            -width 90 -height 31] -expand yes -fill both -side left
    bind $mff.f.txt <FocusIn> { 
        %W configure -state disable 
        %W mark set insert 0.0
    }
    bind $mff.f.txt <Up>   { %W yview scroll -1 units }
    bind $mff.f.txt <Down> { %W yview scroll +1 units }
    $mff.f.txt tag configure <h1> -font [option get $mff.f.txt bigFont $mff.f.txt] \
                                -foreground blue4 -justify center
    $mff.f.txt tag configure <h2> -foreground blue3 -justify center
    $mff.f.txt tag configure <l1> -background lightblue2
    $mff.f.txt tag configure <u>  -underline yes
    $mff.f.txt tag configure <ra> -relief raised -borderwidth 1  -offset 1
    $mff.f.txt tag configure <su> -relief sunken -borderwidth 1 -offset 1
    $mff.f.txt tag configure <gr> -relief groove -borderwidth 2 -offset 2
    $mff.f.txt tag configure <ri> -relief ridge -borderwidth 2 -offset 2
    
    set lst [Help:Read $w $target]
    set lnglast [string length " $msg "]
    set row 1
    foreach l $lst {
        Lindexes [Help:Txt:Tag $row $l] str ltags
        set lng [string length $str]
        if { $lng > $lnglast } { set lnglast $lng }
        $mff.f.txt insert end "$str\n"
        foreach { tg x1 x2 } $ltags {
            eval $mff.f.txt tag add $tg $x1 $x2
        }
        incr row
    }
    if { $row > 31 }     { set row 31 }
    if { $lnglast > 90 } { set lnglast 90 }
    $mff.f.txt configure -width $lnglast -height $row
    $tb.bb1 itemconfigure 0 -command [list Help:Print $w $lst]
    
    bind $w <Escape> "$tb.bb2 invoke 0"
    bind $w <F4>     "$tb.bb1 invoke 0"
    focus $mff.f.txt
    BWidget::place $w 0 0 center
    wm deiconify $w
}

proc Help:Print { w lst } {
    global HELP MAIN
    if ![MessageDlg $w.dlg -buttons { } -icon question -type user -message "    $MAIN(Printer) ?" -default 0] {
        BeforeQuery HELP "   ... "
        set i 0
        set fd [open "|$MAIN(Printcmd) -L$MAIN(Printer)" w]
        foreach l $lst {
            puts $fd $l
            ProgressIncr HELP(progress) [incr i] 4
        }
        close $fd
        AfterQuery HELP
    }
}

proc Help:Read { w target } {
    global HELP
    
    set lst {}
    set i 0
    BeforeQuery HELP "   ... "
    set fd [dopen $w help.msg r]
    set flg 0
    while { ![eof $fd] } {
        set str [gets $fd]
        if { $flg && [regexp -- "^\\..*" $str ] } break
        if $flg {
            lappend lst $str
        }
        if [regexp -- "^\\.$target$" $str ] { set flg 1 }
        ProgressIncr HELP(progress) [incr i] 5
    }    
    close $fd
    AfterQuery HELP
    return $lst
}

proc Help:Txt:Tag { row str } {
    set ls ""
    set ltags ""
    regsub -all -- "( )|(<\[^<>]*>)" $str " {&} " str
    foreach i $str {
        if [regexp -- "<.*>" $i] {
            if [IsBe tag$i] {
                unset tag$i
                lappend ltags $row.[lmath_sum [list_lengths $ls]]
            } else {
                set tag$i 1
                lappend ltags $i $row.[lmath_sum [list_lengths $ls]]
            }
        } else {
            lappend ls $i
        }
    }

    set lsr ""
    for { set i 0 } { $i < [llength $ls] } { incr i } {
        append lsr [lindex $ls $i]
    }
    return [list $lsr $ltags]
}
